#include "buzzer.h"

//   
volatile static int g_nNoteT = 0;
//   
volatile static int g_nNoteP = 0;
//  
volatile static int g_nNoteO = 0;
//
volatile static char const *g_pMelody = 0;
// 
volatile static bool g_bRepeat = true;
//    
volatile static int g_nNewFreq = 0;


// -
void buzzer_init_timers() {
	// 3 (  )
	// , OC3A    
	TCCR3A = 0x00 | 1<<COM3A0 | 1<<WGM31| 1<<WGM30; 
	TCCR3B = 0x00 | 1<<WGM33| 1<<WGM32; //0x00 | 1<<WGM32;
 	
	// 0 (  )
	//CTC,  128,   125
	//16000000 / 128 = 125000, 125000/125 = 1000
	//   
	TCCR0 = 0x00 | 1<<WGM01 | 1<<CS02 | 1<<CS00;	
	OCR0 = 124;	
	TIMSK |= 1<<OCIE0;
}

//  
void go_buzz(int nFreq) {	
	// 
	int nCounter;	
	int nTimerState = TCCR3B;
	//       
	if(nTimerState & 1<<CS31) {			
		/* -    		 
		//   
		OCR3A = nFreq;
		//  "" 
		nCounter = TCNT3;
		if(nCounter >= nFreq) {
			TCNT3 = 0x00;
		}
		*/		
		//     
		g_nNewFreq = nFreq;
		//     3
		ETIMSK |= 1<<TOIE3;
	} else {
		//   
		OCR3A = nFreq;
		//    8
		TCCR3B |= 1<<CS31;
	}
}

//   
void buzzer_change_freq() {
	// 
	int nCounter;
	if(g_nNewFreq != 0) {		
		//   
		OCR3A = g_nNewFreq;		
		//  ""  
		//( - !    )
		nCounter = TCNT3;
		if(nCounter >= g_nNewFreq) {
			TCNT3 = 0x00;
		}
		//  
		g_nNewFreq = 0;
		//     3
		ETIMSK &= ~(1<<TOIE3);
	}
}

//  
void stop_buzz() {	
	//  ,  
	TCCR3B &= ~(1<<CS31);
	//     
	TCNT3 = 0x00;
	//C   
	PORTC &= ~(1<<3);
}


//  
void buzzer_play(const char const *pMelody, bool bRepeat) {
	//  
	g_nNoteO = 0;
	g_nNoteT = 0;
	g_pMelody = pMelody;
	g_bRepeat = bRepeat;
}

//  
void buzzer_stop() {
	// 
	stop_buzz();
	//  
	g_nNoteO = 0;
	g_nNoteT = 0;
	g_pMelody = 0;
	g_bRepeat = false;
}

//  
void buzzer_tick() {
	// 
	int nFreq = 0;
	int nFreqNext = 0;
	bool bPlay = false;
	// ,  
	if(g_pMelody != 0) {
		// ?
		if(g_nNoteT <= 0) {
			//      
			if(g_nNoteP > 0) {
				// 
				stop_buzz();
				//    
				g_nNoteT += g_nNoteP;
				g_nNoteP = 0;
				//
				return;
			}
			// ,    
			nFreq = parse_note(g_nNoteO);
			//  
			if(nFreq < 0) {
				if(g_bRepeat) {
					// 
					g_nNoteO = 0;
				} else {
					// 
					g_pMelody = 0;
					// 
					stop_buzz();
				}				
				return;
			}
			//    
			while(g_pMelody[++g_nNoteO] != ':') {}
			//  
			bPlay = false;
			while(!bPlay) {
				switch(g_pMelody[++g_nNoteO]) {
					case '1': case '2': case '3': case '4': case '5':
					case '6': case '7': case '8': case '9':	case '0':
						//  10-  
						if(g_nNoteT != 0) g_nNoteT *= 10;
						//   10- 
						g_nNoteT += g_pMelody[g_nNoteO] - 48;
						break;
					case 0:
						//  
						if(g_pMelody[g_nNoteO-1] != 's') {
							//     
							//   ,   !
							g_nNoteT = 1000 / g_nNoteT;
						}
						if(g_nNoteT == 0) {
							if(g_bRepeat) {
								// 
								g_nNoteO = 0;
							} else {
								// 
								g_nNoteP = 0;
								g_pMelody = 0;
							}								
							// 
							stop_buzz();
							return;
						} else {
							//  
							bPlay = true;
						}
						break;
					case 's':
						break;
					case '-':
						//  
						if(g_pMelody[g_nNoteO-1] != 's') {
							//     
							//   ,   !
							g_nNoteT = 1000 / g_nNoteT;
						}
						//   
						++g_nNoteO;
						//  
						bPlay = true;
						break;															
					case ' ':
						//  
						if(g_pMelody[g_nNoteO-1] != 's') {
							//     
							//   ,   !
							g_nNoteT = 1000 / g_nNoteT;
						}
						//     
						g_nNoteP = g_nNoteT>>3;
						g_nNoteT -= g_nNoteP;
						//   
						++g_nNoteO;
						//  
						bPlay = true;
						break;
					case '>':
						//  
						if(g_pMelody[g_nNoteO-1] != 's') {
							//     
							//   ,   !
							g_nNoteT = 1000 / g_nNoteT;
						}
						// ,    
						nFreqNext = parse_note(++g_nNoteO);
						//      
						if(nFreq <= 0) {
							if(g_bRepeat) {
								// 
								g_nNoteO = 0;
							} else {
								// 
								g_pMelody = 0;
							}
							// 
							stop_buzz();
							return;
						}
						//  
						bPlay = true;
						break;			
					default:
						// 
						stop_buzz();
						return;
				}
			}
			//    
			if(nFreq > 0) {
				// 
				go_buzz(nFreq);
			} else {
				// 
				stop_buzz();
			}
		} else {
			//   
			--g_nNoteT;	
		}				
	}	
}


//    
int parse_note(int nNoteO) {
	// 
	float fFreq = 0;
	int nFreq = 0;
	
	//C 
	//    
	switch(g_pMelody[nNoteO]) {
		case 0:
			return -1;
		case 'C': 	
			if(g_pMelody[nNoteO+1] != '#') {
				fFreq = 261.63; //
			} else {
				fFreq =	277,18; // 
				++nNoteO;
			}
			break;
		case 'D':
			if(g_pMelody[nNoteO+1] != '#') {
				fFreq = 293,66; //
			} else { 
				fFreq = 311,13; // 
				++nNoteO;
			}
			break;
		case 'E':
			if(g_pMelody[nNoteO+1] != '#') {
				fFreq = 329.63; //
			} else { 
				fFreq = 349.23; //
				++nNoteO;
			}
			break;
		case 'F':
			if(g_pMelody[nNoteO+1] != '#') {
				fFreq = 349.23; //
			} else { 
				fFreq = 369.99; // 
				++nNoteO;
			}
			break;
		case 'G':
			if(g_pMelody[nNoteO+1] != '#') {
				fFreq = 392.00; //
			} else { 
				fFreq = 415.30; // 
				++nNoteO;
			}
			break;
		case 'A':
			if(g_pMelody[nNoteO+1] != '#') {
				fFreq = 440.00; //
			} else { 
				fFreq = 466.16; // 
				++nNoteO;
			}
			break;
		case 'B':
			if(g_pMelody[nNoteO+1] != '#') {
				fFreq = 493.88; //
			} else { 
				fFreq = 523.26; //  
				++nNoteO;
			}
			break;
		default: fFreq = -1; break;
	}
	//    (  )
	if(fFreq > 0) {
		nFreq = F_CPU / ((fFreq * 4/**/)*2*8/**/);
		// ,  
		switch(g_pMelody[++nNoteO]) {
			case '0':
				nFreq <<= 1; 
				++nNoteO;
				break;				
			case '1': 
				++nNoteO; 
				break;				
			case '2':
				nFreq >>= 1; 
				++nNoteO;
				break;
			case '3':
				nFreq >>= 2; 
				++nNoteO;
				break;
			case '4':
				nFreq >>= 3; 
				++nNoteO;
				break;
			default:
				break;
		}
	} else {
		nFreq = 0;
		++nNoteO;
	}
	//   
	if(g_pMelody[nNoteO] != ':') {
		return -1;
	} 	
	//   
	return nFreq;
}
